@@ -30,22 +30,8 @@
 
 G_BEGIN_DECLS
 
-
-#define GST_TYPE_DVDEC \
-  (gst_dvdec_get_type())
-#define GST_DVDEC(obj) \
-  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_DVDEC,GstDVDec))
-#define GST_DVDEC_CLASS(klass) \
-  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_DVDEC,GstDVDecClass))
-#define GST_IS_DVDEC(obj) \
-  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_DVDEC))
-#define GST_IS_DVDEC_CLASS(klass) \
-  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_DVDEC))
-
-
-typedef struct _GstDVDec GstDVDec;
-typedef struct _GstDVDecClass GstDVDecClass;
-
+#define GST_TYPE_DVDEC (gst_dvdec_get_type())
+G_DECLARE_FINAL_TYPE (GstDVDec, gst_dvdec, GST, DVDEC, GstElement)
 
 struct _GstDVDec {
   GstElement     element;
@@ -61,7 +47,6 @@ struct _GstDVDec {
   gboolean       PAL;
   gboolean       interlaced;
   gboolean       wide;
-  gint           frame_len;
 
   /* input caps */
   gboolean       sink_negotiated;
@@ -74,7 +59,6 @@ struct _GstDVDec {
   gboolean       need_par;
 
   /* negotiated output */
-  dv_color_space_t space;
   gint           bpp;
   gboolean       src_negotiated;
   
@@ -86,14 +70,6 @@ struct _GstDVDec {
   gboolean       need_segment;
 };
 
-struct _GstDVDecClass {
-  GstElementClass parent_class;
-};
-
-
-GType gst_dvdec_get_type (void);
-
-
 G_END_DECLS
 
 
