@@ -76,6 +76,10 @@ static void
 filter_scanline_yadif_planar (GstDeinterlaceSimpleMethod * self,
     guint8 * out, const GstDeinterlaceScanlineData * scanlines, guint size);
 
+static void
+filter_scanline_yadif_planar_16bits (GstDeinterlaceSimpleMethod * self,
+    guint8 * out, const GstDeinterlaceScanlineData * scanlines, guint size);
+
 static void
 filter_scanline_yadif_semiplanar (GstDeinterlaceSimpleMethod * self,
     guint8 * out, const GstDeinterlaceScanlineData * scanlines, guint size);
@@ -114,6 +118,24 @@ filter_line_c_planar_mode2 (void *ORC_RESTRICT dst,
     const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
     const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w);
 
+static void
+filter_line_c_planar_mode0_16bits (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
+    const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
+    const void *ORC_RESTRICT ttwo, const void *ORC_RESTRICT btwo,
+    const void *ORC_RESTRICT tptwo, const void *ORC_RESTRICT bptwo,
+    const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
+    const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w);
+
+static void
+filter_line_c_planar_mode2_16bits (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
+    const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
+    const void *ORC_RESTRICT ttwo, const void *ORC_RESTRICT btwo,
+    const void *ORC_RESTRICT tptwo, const void *ORC_RESTRICT bptwo,
+    const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
+    const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w);
+
 static void (*filter_mode2) (void *ORC_RESTRICT dst,
     const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
     const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
@@ -130,6 +152,21 @@ static void (*filter_mode0) (void *ORC_RESTRICT dst,
     const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
     const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w);
 
+static void (*filter_mode2_16bits) (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
+    const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
+    const void *ORC_RESTRICT ttwo, const void *ORC_RESTRICT btwo,
+    const void *ORC_RESTRICT tptwo, const void *ORC_RESTRICT bptwo,
+    const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
+    const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w);
+
+static void (*filter_mode0_16bits) (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
+    const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
+    const void *ORC_RESTRICT ttwo, const void *ORC_RESTRICT btwo,
+    const void *ORC_RESTRICT tptwo, const void *ORC_RESTRICT bptwo,
+    const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
+    const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w);
 
 static void
 copy_scanline (GstDeinterlaceSimpleMethod * self, guint8 * out,
@@ -166,6 +203,9 @@ static void
   dism_class->copy_scanline_bgr = copy_scanline;
   dism_class->copy_scanline_nv12 = copy_scanline;
   dism_class->copy_scanline_nv21 = copy_scanline;
+  dism_class->copy_scanline_planar_y_16bits = copy_scanline;
+  dism_class->copy_scanline_planar_u_16bits = copy_scanline;
+  dism_class->copy_scanline_planar_v_16bits = copy_scanline;
 
   dism_class->interpolate_scanline_planar_y = filter_scanline_yadif_planar;
   dism_class->interpolate_scanline_planar_u = filter_scanline_yadif_planar;
@@ -182,6 +222,12 @@ static void
   dism_class->interpolate_scanline_bgr = filter_scanline_yadif_packed_3;
   dism_class->interpolate_scanline_nv12 = filter_scanline_yadif_semiplanar;
   dism_class->interpolate_scanline_nv21 = filter_scanline_yadif_semiplanar;
+  dism_class->interpolate_scanline_planar_y_16bits =
+      filter_scanline_yadif_planar_16bits;
+  dism_class->interpolate_scanline_planar_u_16bits =
+      filter_scanline_yadif_planar_16bits;
+  dism_class->interpolate_scanline_planar_v_16bits =
+      filter_scanline_yadif_planar_16bits;
 }
 
 #define FFABS(a) ABS(a)
@@ -190,18 +236,19 @@ static void
 #define FFMAX3(a,b,c) FFMAX(FFMAX(a,b),c)
 #define FFMIN3(a,b,c) FFMIN(FFMIN(a,b),c)
 
-#define CHECK(j1, j2, j3)\
-    {   int score = FFABS(stzero[x - j1] - sbzero[x - j2])\
-                  + FFABS(stzero[x + j3] - sbzero[x - j3])\
-                  + FFABS(stzero[x + j1] - sbzero[x + j2]);\
+
+#define CHECK(j)\
+    {   int score = FFABS(stzero[x - colors2 + j] - sbzero[x - colors2 - j])\
+                  + FFABS(stzero[x  + j] - sbzero[x  - j])\
+                  + FFABS(stzero[x + colors2 + j] - sbzero[x + colors2 - j]);\
         if (score < spatial_score) {\
             spatial_score= score;\
-            spatial_pred= (stzero[x + j3] + sbzero[x - j3])>>1;\
+            spatial_pred= (stzero[x  + j] + sbzero[x - j])>>1;\
 
 /* The is_not_edge argument here controls when the code will enter a branch
  * which reads up to and including x-3 and x+3. */
 
-#define FILTER(start, end, is_not_edge) \
+#define FILTER(start, end, is_not_edge) G_STMT_START { \
     for (x = start;  x < end; x++) { \
         int c = stzero[x]; \
         int d = (smone[x] + smp[x])>>1; \
@@ -219,14 +266,8 @@ static void
         if (is_not_edge) {\
             int spatial_score = FFABS(stzero[x-colors2] - sbzero[x-colors2]) + FFABS(c-e) \
                               + FFABS(stzero[x+colors2] - sbzero[x+colors2]); \
-            int twice_colors2 = colors2 << 1; \
-            int minus_colors2 = -colors2; \
-            int thrice_colors2 = colors2 * 3; \
-            int minus2_colors2 = colors2 * -2; \
-            CHECK(0, twice_colors2, minus_colors2) \
-              CHECK(-colors2, thrice_colors2, minus2_colors2) }} }} \
-            CHECK(twice_colors2, 0, colors2) \
-              CHECK(thrice_colors2, minus_colors2, twice_colors2) }} }} \
+            CHECK(-1 * colors2) CHECK(-2 * colors2) }} }} \
+            CHECK(colors2) CHECK(2 * colors2) }} }} \
         }\
  \
         if (!(mode&2)) { \
@@ -245,7 +286,8 @@ static void
  \
         sdst[x] = spatial_pred; \
  \
-    }
+    } \
+} G_STMT_END
 
 ALWAYS_INLINE static void
 filter_line_c (guint8 * sdst, const guint8 * stzero, const guint8 * sbzero,
@@ -262,7 +304,7 @@ filter_line_c (guint8 * sdst, const guint8 * stzero, const guint8 * sbzero,
    * This allows the FILTER macro to be
    * called so that it processes all the pixels normally.  A constant value of
    * true for is_not_edge lets the compiler ignore the if statement. */
-  FILTER (start, end, 1)
+  FILTER (start, end, 1);
 }
 
 #define MAX_ALIGN 8
@@ -300,7 +342,43 @@ filter_line_c_planar (void *ORC_RESTRICT dst, const void *ORC_RESTRICT tzero,
    * This allows the FILTER macro to be
    * called so that it processes all the pixels normally.  A constant value of
    * true for is_not_edge lets the compiler ignore the if statement. */
-  FILTER (start, end, 1)
+  FILTER (start, end, 1);
+}
+
+ALWAYS_INLINE static void
+filter_line_c_planar_16bits (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero,
+    const void *ORC_RESTRICT bzero, const void *ORC_RESTRICT mone,
+    const void *ORC_RESTRICT mp, const void *ORC_RESTRICT ttwo,
+    const void *ORC_RESTRICT btwo, const void *ORC_RESTRICT tptwo,
+    const void *ORC_RESTRICT bptwo, const void *ORC_RESTRICT ttone,
+    const void *ORC_RESTRICT ttp, const void *ORC_RESTRICT bbone,
+    const void *ORC_RESTRICT bbp, int w, int mode)
+{
+  int x;
+  const int start = 0;
+  const int colors = 1;
+  const int y_alternates_every = 0;
+  const int end = w;
+  guint16 *sdst = (guint16 *) dst + 3;
+  guint16 *stzero = (guint16 *) tzero + 3;
+  guint16 *sbzero = (guint16 *) bzero + 3;
+  guint16 *smone = (guint16 *) mone + 3;
+  guint16 *smp = (guint16 *) mp + 3;
+  guint16 *sttwo = (guint16 *) ttwo + 3;
+  guint16 *sbtwo = (guint16 *) btwo + 3;
+  guint16 *stptwo = (guint16 *) tptwo + 3;
+  guint16 *sbptwo = (guint16 *) bptwo + 3;
+  guint16 *sttone = (guint16 *) ttone + 3;
+  guint16 *sttp = (guint16 *) ttp + 3;
+  guint16 *sbbone = (guint16 *) bbone + 3;
+  guint16 *sbbp = (guint16 *) bbp + 3;
+  /* The function is called for processing the middle
+   * pixels of each line, excluding 3 at each end.
+   * This allows the FILTER macro to be
+   * called so that it processes all the pixels normally.  A constant value of
+   * true for is_not_edge lets the compiler ignore the if statement. */
+  FILTER (start, end, 1);
 }
 
 ALWAYS_INLINE G_GNUC_UNUSED static void
@@ -329,6 +407,32 @@ filter_line_c_planar_mode2 (void *ORC_RESTRICT dst,
       ttone, ttp, bbone, bbp, w, 2);
 }
 
+ALWAYS_INLINE G_GNUC_UNUSED static void
+filter_line_c_planar_mode0_16bits (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
+    const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
+    const void *ORC_RESTRICT ttwo, const void *ORC_RESTRICT btwo,
+    const void *ORC_RESTRICT tptwo, const void *ORC_RESTRICT bptwo,
+    const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
+    const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w)
+{
+  filter_line_c_planar_16bits (dst, tzero, bzero, mone, mp, ttwo, btwo, tptwo,
+      bptwo, ttone, ttp, bbone, bbp, w, 0);
+}
+
+ALWAYS_INLINE G_GNUC_UNUSED static void
+filter_line_c_planar_mode2_16bits (void *ORC_RESTRICT dst,
+    const void *ORC_RESTRICT tzero, const void *ORC_RESTRICT bzero,
+    const void *ORC_RESTRICT mone, const void *ORC_RESTRICT mp,
+    const void *ORC_RESTRICT ttwo, const void *ORC_RESTRICT btwo,
+    const void *ORC_RESTRICT tptwo, const void *ORC_RESTRICT bptwo,
+    const void *ORC_RESTRICT ttone, const void *ORC_RESTRICT ttp,
+    const void *ORC_RESTRICT bbone, const void *ORC_RESTRICT bbp, int w)
+{
+  filter_line_c_planar_16bits (dst, tzero, bzero, mone, mp, ttwo, btwo, tptwo,
+      bptwo, ttone, ttp, bbone, bbp, w, 2);
+}
+
 ALWAYS_INLINE static void
 filter_edges (guint8 * sdst, const guint8 * stzero, const guint8 * sbzero,
     const guint8 * smone, const guint8 * smp, const guint8 * sttwo,
@@ -343,9 +447,42 @@ filter_edges (guint8 * sdst, const guint8 * stzero, const guint8 * sbzero,
 
   /* Only edge pixels need to be processed here.  A constant value of false
    * for is_not_edge should let the compiler ignore the whole branch. */
-  FILTER (0, border, 0)
-      FILTER (w - edge, w - border, 1)
-      FILTER (w - border, w, 0)
+  FILTER (0, border, 0);
+  FILTER (w - edge, w - border, 1);
+  FILTER (w - border, w, 0);
+}
+
+ALWAYS_INLINE static void
+filter_edges_16bits (guint8 * sdst_, const guint8 * stzero_,
+    const guint8 * sbzero_,
+    const guint8 * smone_, const guint8 * smp_, const guint8 * sttwo_,
+    const guint8 * sbtwo_, const guint8 * stptwo_, const guint8 * sbptwo_,
+    const guint8 * sttone_, const guint8 * sttp_, const guint8 * sbbone_,
+    const guint8 * sbbp_, int w, int colors, int y_alternates_every,
+    int mode, const int bpp)
+{
+  int x;
+  const int edge = colors * (MAX_ALIGN / bpp);
+  const int border = 3 * colors;
+  guint16 *sdst = (guint16 *) sdst_;
+  guint16 *stzero = (guint16 *) stzero_;
+  guint16 *sbzero = (guint16 *) sbzero_;
+  guint16 *smone = (guint16 *) smone_;
+  guint16 *smp = (guint16 *) smp_;
+  guint16 *sttwo = (guint16 *) sttwo_;
+  guint16 *sbtwo = (guint16 *) sbtwo_;
+  guint16 *stptwo = (guint16 *) stptwo_;
+  guint16 *sbptwo = (guint16 *) sbptwo_;
+  guint16 *sttone = (guint16 *) sttone_;
+  guint16 *sttp = (guint16 *) sttp_;
+  guint16 *sbbone = (guint16 *) sbbone_;
+  guint16 *sbbp = (guint16 *) sbbp_;
+
+  /* Only edge pixels need to be processed here.  A constant value of false
+   * for is_not_edge should let the compiler ignore the whole branch. */
+  FILTER (0, border, 0);
+  FILTER (w - edge, w - border, 1);
+  FILTER (w - border, w, 0);
 }
 
 static void
@@ -455,9 +592,49 @@ filter_scanline_yadif_planar (GstDeinterlaceSimpleMethod * self,
         (void *) s.bbp, w - edge);
 }
 
+ALWAYS_INLINE static void
+filter_scanline_yadif_planar_16bits (GstDeinterlaceSimpleMethod * self,
+    guint8 * out, const GstDeinterlaceScanlineData * s_orig, guint size)
+{
+  guint8 *dst = out;
+  const int bpp = 2;
+  int w = size / bpp;
+  int edge = MAX_ALIGN / bpp;
+  GstDeinterlaceScanlineData s = *s_orig;
+
+  int mode = (s.tt1 == NULL || s.bb1 == NULL || s.ttp == NULL
+      || s.bbp == NULL) ? 2 : 0;
+
+  /* When starting up, some data might not yet be available, so use the current frame */
+  if (s.m1 == NULL)
+    s.m1 = s.mp;
+  if (s.tt1 == NULL)
+    s.tt1 = s.ttp;
+  if (s.bb1 == NULL)
+    s.bb1 = s.bbp;
+  if (s.t2 == NULL)
+    s.t2 = s.tp2;
+  if (s.b2 == NULL)
+    s.b2 = s.bp2;
+
+  filter_edges_16bits (dst, s.t0, s.b0, s.m1, s.mp, s.t2, s.b2, s.tp2, s.bp2,
+      s.tt1, s.ttp, s.bb1, s.bbp, w, 1, 0, mode, bpp);
+  if (mode == 0)
+    filter_mode0_16bits (dst, (void *) s.t0, (void *) s.b0, (void *) s.m1,
+        (void *) s.mp, (void *) s.t2, (void *) s.b2, (void *) s.tp2,
+        (void *) s.bp2, (void *) s.tt1, (void *) s.ttp, (void *) s.bb1,
+        (void *) s.bbp, w - edge);
+  else
+    filter_mode2_16bits (dst, (void *) s.t0, (void *) s.b0, (void *) s.m1,
+        (void *) s.mp, (void *) s.t2, (void *) s.b2, (void *) s.tp2,
+        (void *) s.bp2, (void *) s.tt1, (void *) s.ttp, (void *) s.bb1,
+        (void *) s.bbp, w - edge);
+}
+
 static void
 gst_deinterlace_method_yadif_init (GstDeinterlaceMethodYadif * self)
 {
+  /* TODO: add asm support for high bitdepth */
 #if (defined __x86_64__ || defined _M_X64) && defined HAVE_NASM
   if (
 #  if defined HAVE_ORC
@@ -471,16 +648,22 @@ gst_deinterlace_method_yadif_init (GstDeinterlaceMethodYadif * self)
     GST_DEBUG ("SSSE3 optimization enabled");
     filter_mode0 = gst_yadif_filter_line_mode0_ssse3;
     filter_mode2 = gst_yadif_filter_line_mode2_ssse3;
+    filter_mode0_16bits = filter_line_c_planar_mode0_16bits;
+    filter_mode2_16bits = filter_line_c_planar_mode2_16bits;
   } else {
     GST_DEBUG ("SSE2 optimization enabled");
     filter_mode0 = gst_yadif_filter_line_mode0_sse2;
     filter_mode2 = gst_yadif_filter_line_mode2_sse2;
+    filter_mode0_16bits = filter_line_c_planar_mode0_16bits;
+    filter_mode2_16bits = filter_line_c_planar_mode2_16bits;
   }
 #else
   {
     GST_DEBUG ("SSE optimization disabled");
     filter_mode0 = filter_line_c_planar_mode0;
     filter_mode2 = filter_line_c_planar_mode2;
+    filter_mode0_16bits = filter_line_c_planar_mode0_16bits;
+    filter_mode2_16bits = filter_line_c_planar_mode2_16bits;
   }
 #endif
 }
