@@ -39,6 +39,16 @@ GST_DEBUG_CATEGORY_STATIC (rtph265depay_debug);
  * expressed a restriction or preference via caps */
 #define DEFAULT_STREAM_FORMAT GST_H265_STREAM_FORMAT_BYTESTREAM
 #define DEFAULT_ACCESS_UNIT   FALSE
+#define DEFAULT_WAIT_FOR_KEYFRAME FALSE
+#define DEFAULT_REQUEST_KEYFRAME FALSE
+
+enum
+{
+  PROP_0,
+  PROP_WAIT_FOR_KEYFRAME,
+  PROP_REQUEST_KEYFRAME,
+};
+
 
 /* 3 zero bytes syncword */
 static const guint8 sync_bytes[] = { 0, 0, 0, 1 };
@@ -119,6 +129,44 @@ static void gst_rtp_h265_depay_push (GstRtpH265Depay * rtph265depay,
     GstBuffer * outbuf, gboolean keyframe, GstClockTime timestamp,
     gboolean marker);
 
+static void
+gst_rtp_h265_depay_set_property (GObject * object, guint prop_id,
+    const GValue * value, GParamSpec * pspec)
+{
+  GstRtpH265Depay *self = GST_RTP_H265_DEPAY (object);
+
+  switch (prop_id) {
+    case PROP_WAIT_FOR_KEYFRAME:
+      self->wait_for_keyframe = g_value_get_boolean (value);
+      break;
+    case PROP_REQUEST_KEYFRAME:
+      self->request_keyframe = g_value_get_boolean (value);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+}
+
+static void
+gst_rtp_h265_depay_get_property (GObject * object, guint prop_id,
+    GValue * value, GParamSpec * pspec)
+{
+  GstRtpH265Depay *self = GST_RTP_H265_DEPAY (object);
+
+  switch (prop_id) {
+    case PROP_WAIT_FOR_KEYFRAME:
+      g_value_set_boolean (value, self->wait_for_keyframe);
+      break;
+    case PROP_REQUEST_KEYFRAME:
+      g_value_set_boolean (value, self->request_keyframe);
+      break;
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      break;
+  }
+}
+
 
 static void
 gst_rtp_h265_depay_class_init (GstRtpH265DepayClass * klass)
@@ -132,6 +180,37 @@ gst_rtp_h265_depay_class_init (GstRtpH265DepayClass * klass)
   gstrtpbasedepayload_class = (GstRTPBaseDepayloadClass *) klass;
 
   gobject_class->finalize = gst_rtp_h265_depay_finalize;
+  gobject_class->set_property = gst_rtp_h265_depay_set_property;
+  gobject_class->get_property = gst_rtp_h265_depay_get_property;
+
+  /**
+   * GstRtpH265Depay:wait-for-keyframe:
+   *
+   * Wait for the next keyframe after packet loss.
+   *
+   * Note: currently this only has an effect if outputting access units.
+   *
+   * Since: 1.26
+   */
+  g_object_class_install_property (gobject_class, PROP_WAIT_FOR_KEYFRAME,
+      g_param_spec_boolean ("wait-for-keyframe", "Wait for Keyframe",
+          "Wait for the next keyframe after packet loss, meaningful only when "
+          "outputting access units",
+          DEFAULT_WAIT_FOR_KEYFRAME,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
+  /**
+   * GstRtpH265Depay:request-keyframe:
+   *
+   * Request new keyframe when packet loss is detected.
+   *
+   * Since: 1.26
+   */
+  g_object_class_install_property (gobject_class, PROP_REQUEST_KEYFRAME,
+      g_param_spec_boolean ("request-keyframe", "Request Keyframe",
+          "Request new keyframe when packet loss is detected",
+          DEFAULT_REQUEST_KEYFRAME,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
   gst_element_class_add_static_pad_template (gstelement_class,
       &gst_rtp_h265_depay_src_template);
@@ -168,6 +247,8 @@ gst_rtp_h265_depay_init (GstRtpH265Depay * rtph265depay)
       (GDestroyNotify) gst_buffer_unref);
   rtph265depay->pps = g_ptr_array_new_with_free_func (
       (GDestroyNotify) gst_buffer_unref);
+  rtph265depay->wait_for_keyframe = DEFAULT_WAIT_FOR_KEYFRAME;
+  rtph265depay->request_keyframe = DEFAULT_REQUEST_KEYFRAME;
 }
 
 static void
@@ -175,6 +256,8 @@ gst_rtp_h265_depay_reset (GstRtpH265Depay * rtph265depay, gboolean hard)
 {
   gst_adapter_clear (rtph265depay->adapter);
   rtph265depay->wait_start = TRUE;
+  rtph265depay->waiting_for_keyframe = rtph265depay->wait_for_keyframe;
+  rtph265depay->requesting_keyframe = FALSE;
   gst_adapter_clear (rtph265depay->picture_adapter);
   rtph265depay->picture_start = FALSE;
   rtph265depay->last_keyframe = FALSE;
@@ -420,6 +503,7 @@ gst_rtp_h265_set_src_caps (GstRtpH265Depay * rtph265depay)
     GstBitReader br;
     guint32 tmp;
     guint8 tmp8 = 0;
+    gsize src_offset, dst_offset;
     guint32 max_sub_layers_minus1, temporal_id_nesting_flag, chroma_format_idc,
         bit_depth_luma_minus8, bit_depth_chroma_minus8,
         min_spatial_segmentation_idc;
@@ -461,7 +545,42 @@ gst_rtp_h265_set_src_caps (GstRtpH265Depay * rtph265depay)
     max_sub_layers_minus1 = ((nalmap.data[2]) >> 1) & 0x07;
     temporal_id_nesting_flag = nalmap.data[2] & 0x01;
 
-    gst_bit_reader_init (&br, nalmap.data + 15, nalmap.size - 15);
+    /* HEVCDecoderConfigurationVersion = 1 */
+    data[0] = 1;
+
+    /* Copy from profile_tier_level (Rec. ITU-T H.265 (04/2013) section 7.3.3
+     *
+     * profile_space | tier_flat | profile_idc |
+     * profile_compatibility_flags | constraint_indicator_flags |
+     * level_idc | progressive_source_flag | interlaced_source_flag
+     * non_packed_constraint_flag | frame_only_constraint_flag
+     * reserved_zero_44bits | level_idc */
+
+    guint n_zero_bytes = 0;
+    // 16 bit nal header and sps-vps id.
+    for (src_offset = 3, dst_offset = 1; dst_offset < 13; src_offset++) {
+      if (src_offset >= nalmap.size)
+        // SPS not long enough
+        return FALSE;
+
+      if (n_zero_bytes == 2 && nalmap.data[src_offset] == 0x03) {
+        n_zero_bytes = 0;
+        // ignore emulation prevention bytes: [0x00, 0x00, 0x03]
+        continue;
+      }
+
+      if (nalmap.data[src_offset] == 0x00)
+        n_zero_bytes++;
+      else
+        n_zero_bytes = 0;
+
+      data[dst_offset] = nalmap.data[src_offset];
+      dst_offset++;
+    }
+
+    // FIXME: does not parse any sub layers
+    gst_bit_reader_init (&br, nalmap.data + src_offset,
+        nalmap.size - src_offset);
 
     gst_rtp_read_golomb (&br, &tmp);    /* sps_seq_parameter_set_id */
     gst_rtp_read_golomb (&br, &chroma_format_idc);      /* chroma_format_idc */
@@ -487,23 +606,6 @@ gst_rtp_h265_set_src_caps (GstRtpH265Depay * rtph265depay)
         "Ignoring min_spatial_segmentation for now (assuming zero)");
 
     min_spatial_segmentation_idc = 0;   /* NOTE - we ignore this for now, but in a perfect world, we should continue parsing to obtain the real value */
-
-    gst_buffer_unmap (g_ptr_array_index (rtph265depay->sps, 0), &nalmap);
-
-    /* HEVCDecoderConfigurationVersion = 1 */
-    data[0] = 1;
-
-    /* Copy from profile_tier_level (Rec. ITU-T H.265 (04/2013) section 7.3.3
-     *
-     * profile_space | tier_flat | profile_idc |
-     * profile_compatibility_flags | constraint_indicator_flags |
-     * level_idc | progressive_source_flag | interlaced_source_flag
-     * non_packed_constraint_flag | frame_only_constraint_flag
-     * reserved_zero_44bits | level_idc */
-    gst_buffer_map (g_ptr_array_index (rtph265depay->sps, 0), &nalmap,
-        GST_MAP_READ);
-    for (i = 0; i < 12; i++)
-      data[i + 1] = nalmap.data[i];
     gst_buffer_unmap (g_ptr_array_index (rtph265depay->sps, 0), &nalmap);
 
     /* min_spatial_segmentation_idc */
@@ -1171,30 +1273,36 @@ gst_rtp_h265_depay_handle_nal (GstRtpH265Depay * rtph265depay, GstBuffer * nal,
   if (rtph265depay->merge) {
     gboolean start = FALSE, complete = FALSE;
 
+    if (NAL_TYPE_IS_CODED_SLICE_SEGMENT (nal_type)) {
+      /* A NAL unit (X) ends an access unit if the next-occurring VCL NAL unit (Y) has the high-order bit of the first byte after its NAL unit header equal to 1 */
+      start = TRUE;
+      if (((map.data[6] >> 7) & 0x01) == 1) {
+        complete = TRUE;
+      }
+    } else if ((nal_type >= 32 && nal_type <= 35)
+        || nal_type == 39 || (nal_type >= 41 && nal_type <= 44)
+        || (nal_type >= 48 && nal_type <= 55)) {
+      /* VPS, SPS, PPS, SEI, ... terminate an access unit */
+      complete = TRUE;
+    }
+    GST_DEBUG_OBJECT (depayload, "start %d, complete %d", start, complete);
+
     /* marker bit isn't mandatory so in the following code we try to detect
      * an AU boundary (see H.265 spec section 7.4.2.4.4) */
     if (!marker) {
-      if (NAL_TYPE_IS_CODED_SLICE_SEGMENT (nal_type)) {
-        /* A NAL unit (X) ends an access unit if the next-occurring VCL NAL unit (Y) has the high-order bit of the first byte after its NAL unit header equal to 1 */
-        start = TRUE;
-        if (((map.data[6] >> 7) & 0x01) == 1) {
-          complete = TRUE;
-        }
-      } else if ((nal_type >= 32 && nal_type <= 35)
-          || nal_type == 39 || (nal_type >= 41 && nal_type <= 44)
-          || (nal_type >= 48 && nal_type <= 55)) {
-        /* VPS, SPS, PPS, SEI, ... terminate an access unit */
-        complete = TRUE;
-      }
-      GST_DEBUG_OBJECT (depayload, "start %d, complete %d", start, complete);
-
       if (complete && rtph265depay->picture_start)
         outbuf = gst_rtp_h265_complete_au (rtph265depay, &out_timestamp,
             &out_keyframe);
     }
+
     /* add to adapter */
     gst_buffer_unmap (nal, &map);
 
+    if (!rtph265depay->picture_start && start && out_keyframe) {
+      rtph265depay->waiting_for_keyframe = FALSE;
+      rtph265depay->requesting_keyframe = FALSE;
+    }
+
     GST_DEBUG_OBJECT (depayload, "adding NAL to picture adapter");
     gst_adapter_push (rtph265depay->picture_adapter, nal);
     rtph265depay->last_ts = in_timestamp;
@@ -1212,8 +1320,26 @@ gst_rtp_h265_depay_handle_nal (GstRtpH265Depay * rtph265depay, GstBuffer * nal,
   }
 
   if (outbuf) {
-    gst_rtp_h265_depay_push (rtph265depay, outbuf, out_keyframe, out_timestamp,
-        marker);
+    /* Request a new keyframe if we are waiting for one */
+    if (rtph265depay->waiting_for_keyframe &&
+        !rtph265depay->requesting_keyframe && rtph265depay->request_keyframe) {
+      rtph265depay->requesting_keyframe = TRUE;
+      GST_INFO_OBJECT (depayload, "Requesting keyframe while waiting for one");
+      gst_pad_push_event (GST_RTP_BASE_DEPAYLOAD_SINKPAD (depayload),
+          gst_video_event_new_upstream_force_key_unit (GST_CLOCK_TIME_NONE,
+              TRUE, 0));
+    }
+
+    if (!rtph265depay->waiting_for_keyframe || !rtph265depay->wait_for_keyframe) {
+      GST_INFO_OBJECT (depayload, "Pushing frame %" GST_PTR_FORMAT, outbuf);
+      gst_rtp_h265_depay_push (rtph265depay, outbuf, out_keyframe,
+          out_timestamp, marker);
+    } else {
+      GST_INFO_OBJECT (depayload,
+          "Dropping %" GST_PTR_FORMAT ", we are waiting for a keyframe",
+          outbuf);
+      gst_buffer_unref (outbuf);
+    }
   }
 
   return;
@@ -1261,16 +1387,26 @@ gst_rtp_h265_depay_process (GstRTPBaseDepayload * depayload, GstRTPBuffer * rtp)
 {
   GstRtpH265Depay *rtph265depay;
   GstBuffer *outbuf = NULL;
+  gboolean is_discont;
   guint8 nal_unit_type;
 
   rtph265depay = GST_RTP_H265_DEPAY (depayload);
 
+  if (!rtph265depay->merge) {
+    rtph265depay->waiting_for_keyframe = FALSE;
+    rtph265depay->requesting_keyframe = FALSE;
+  }
+
   /* flush remaining data on discont */
-  if (GST_BUFFER_IS_DISCONT (rtp->buffer)) {
+  is_discont = GST_BUFFER_IS_DISCONT (rtp->buffer);
+  if (is_discont) {
     gst_adapter_clear (rtph265depay->adapter);
     rtph265depay->wait_start = TRUE;
     rtph265depay->current_fu_type = 0;
     rtph265depay->last_fu_seqnum = 0;
+
+    if (rtph265depay->merge)
+      rtph265depay->waiting_for_keyframe = TRUE;
   }
 
   {
@@ -1510,6 +1646,16 @@ gst_rtp_h265_depay_process (GstRTPBaseDepayload * depayload, GstRTPBuffer * rtp)
           /* and assemble in the adapter */
           gst_adapter_push (rtph265depay->adapter, outbuf);
         } else {
+          /* If packet is discont and is not the first one of a FU, then we
+           * can tell it is not part of a keyframe, so request a new one */
+          if (is_discont && rtph265depay->request_keyframe) {
+            GST_INFO_OBJECT (depayload,
+                "discont FU received without start bit. Requesting keyframe.");
+            gst_pad_push_event (GST_RTP_BASE_DEPAYLOAD_SINKPAD (depayload),
+                gst_video_event_new_upstream_force_key_unit
+                (GST_CLOCK_TIME_NONE, TRUE, 0));
+          }
+
           if (rtph265depay->current_fu_type == 0) {
             /* previous FU packet missing start bit? */
             GST_WARNING_OBJECT (rtph265depay, "missing FU start bit on an "
